#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/stat.h>

#include "config.h"

#define BSIZE 512

const char *ok = "+OK";
struct in_addr saddr;

struct hostent *get_hostent(char *host)
{
int c;
struct hostent *hostinfo;
if (host == NULL)
	{
	 fprintf(stderr,"No hostname given !\n");
	 exit(EXIT_FAILURE);
	}
c = *host;
if (isdigit(c))
	{
	 saddr.s_addr = inet_addr(host);
	 hostinfo = gethostbyaddr((char *)&saddr, sizeof(struct in_addr), AF_INET);
	 if (hostinfo == NULL)
		{
		 fprintf(stderr, "Network is unreachable\n");
		 exit(EXIT_FAILURE);
		}
	}
else
	{
	 hostinfo = gethostbyname(host);
	 if (hostinfo == NULL)
		{
		 fprintf(stderr, "%s: Unknown host\n", host);
		 exit(EXIT_FAILURE);
		}
	}
return hostinfo;
}

int start_connection(char *host, unsigned short int port)
{
struct hostent *hostinfo;
struct sockaddr_in addr;
int sfd;
sfd = socket(AF_INET, SOCK_STREAM, 0);
if (sfd < 0)
	{
	 perror("socket: creating socket");
	 exit(EXIT_FAILURE);
	}

hostinfo = get_hostent(host);
addr.sin_family = AF_INET;
addr.sin_port = htons(port);
addr.sin_addr = saddr;
if (connect(sfd, (struct sockaddr *)&addr, sizeof addr) == -1)
	{
	 perror(NULL);
	 exit(EXIT_FAILURE);
	}
printf("Connected to %s.\n", hostinfo->h_name);
return sfd;
}

get_data(int sockfd, char *buffer, int length)
{
while (--length)
	{
	 if (read(sockfd, buffer, 1) == -1)
		{
		 perror("socket: reading data");
		 exit(EXIT_FAILURE);
		}
	 if (*buffer == '\n') break;
	 if (*buffer != '\r') buffer++;
	}
buffer++;
*buffer = 0;
}

send_data(int sockfd, char *buffer, int length)
{
int err;
err = write(sockfd, buffer, length);
if (err == -1)
	{
	 perror("socket: writing data");
	 exit(EXIT_FAILURE);
	}
}

pop_ok(char *buffer)
{
if (strncmp(buffer, ok, 3) != 0)
	{
	 fprintf(stderr, "POP server: error\n");
	 exit(EXIT_FAILURE);
	}
}

int hwmany(char *buffer, int start)
{
int i, hm;
for (i = 0; i < start; i++) buffer++;
hm = strtol(buffer, NULL, 0);
return hm;
}

close_connection(int sockfd)
{
send_data(sockfd, "QUIT\r\n", 6);
close(sockfd);
}

void main(int argc, char **argv)
{
int i, n, dim, fd, sockfd;
char buffer[BSIZE];
char k[BSIZE];
if (argc < 5)
	{
	 if (argc != 1 || read_config(CONFIG, argv[0]) == -1)
		{
		 printf(BANNER);
	 	 printf("usage: %s hostname localusercode remoteusercode password [port]\n", argv[0]);
		 exit(EXIT_FAILURE);
		}
	}
if (argc < 2)
	{
	 exit(EXIT_FAILURE);
	} 
printf(BANNER);
sockfd = start_connection(argv[1], argc < 6 ? POP3_PORT : strtol(argv[5], NULL, 0));
get_data(sockfd, buffer, BSIZE);
if (strncmp(buffer, ok, 3) != 0)
	{
	 fprintf(stderr, "POP server unavailable\n");
	 exit(EXIT_FAILURE);
	}
send_data(sockfd, "USER ", 5);
send_data(sockfd, argv[3], strlen(argv[3]));
send_data(sockfd, "\r\n", 2);
get_data(sockfd, buffer, BSIZE);
pop_ok(buffer);
send_data(sockfd, "PASS ", 5);
send_data(sockfd, argv[4], strlen(argv[4]));
send_data(sockfd, "\r\n", 2);
get_data(sockfd, buffer, BSIZE);
if (strncmp(buffer, ok, 3) != 0)
	{
	 fprintf(stderr, "POP server: invalid usercode or password\n");
	 exit(EXIT_FAILURE);
	}
send_data(sockfd, "STAT\r\n", 6);
get_data(sockfd, buffer, BSIZE);
pop_ok(buffer);
n = hwmany(buffer, 4);
if (n == 0)
	{
	 printf("No mail.\n");
	}
else
	{
	 fd = open(strcat(strcat(SPOOL, "/"), argv[2]), (O_WRONLY | O_APPEND | O_CREAT), (S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP));
	 if (fd == -1)
		{
		 perror("opening spool file");
		 exit(EXIT_FAILURE);
		}
	}	
i = 0;
while (i < n)
	{
	 i++;
	 sprintf(k, "%d", i);
	 send_data(sockfd, "LIST ", 5);
	 send_data(sockfd, k, strlen(k));
	 send_data(sockfd, "\r\n", 2);
	 get_data(sockfd, buffer, BSIZE);
	 pop_ok(buffer);
	 sprintf(k, "%d", n);
	 dim = hwmany(buffer, strlen(k) + 5);
	 sprintf(k, "%d", i); 
	 send_data(sockfd, "RETR ", 5);
	 send_data(sockfd, k, strlen(k));
	 send_data(sockfd, "\r\n", 2);
	 get_data(sockfd, buffer, BSIZE);
	 pop_ok(buffer);
	 retrieve_message(sockfd, dim, fd);
	 get_data(sockfd, buffer, BSIZE);
	 get_data(sockfd, buffer, BSIZE);
#ifdef DELETION
	 send_data(sockfd, "DELE ", 5);
	 send_data(sockfd, k, strlen(k));
	 send_data(sockfd, "\r\n", 2);
	 get_data(sockfd, buffer, BSIZE);
	 pop_ok(buffer);
#endif
	}
fchown(fd, 0, MAIL_GRP);
i = close(fd);
if (i == -1)
	{
	 perror("closing spool file");
	 exit(EXIT_FAILURE);
	}
close_connection(sockfd);
if (n > 0) printf("Done. (%d new message%s)\n", n, n == 1 ? "" : "s");
}
